/* ***** BEGIN LICENSE BLOCK *****
* Version: MPL 1.1/GPL 2.0/LGPL 2.1
*
* The contents of this file are subject to the Mozilla Public License
* Version 1.1 (the "License"); you may not use this file except in
* compliance with the License. You may obtain a copy of the License at
* http://www.mozilla.org/MPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
* for the specific language governing rights and limitations under the
* License.
*
* The Original Code is Accessibar
*
* The Initial Developer of the Original Code is
* Edan Kemelman <edankeme@yahoo.com>
* Portions created by the Initial Developer are Copyright (C) 2004
* the Initial Developer. All Rights Reserved.
*
* Contributor(s):
*    Edan Kemelman <edankeme@yahoo.com>
*
* Alternatively, the contents of this file may be used under the
* terms of the GNU Public License (the "GPL"), in which case the
* provisions of the GPL are applicable instead of those above.
* If you wish to allow use of your version of this file only
* under the terms of the GPL and not to allow others to use your
* version of this file under the MPL, indicate your decision by
* deleting the provisions above and replace them with the notice
* and other provisions required by the GPL.  If you do not delete
* the provisions above, a recipient may use your version of this
* file under either the MPL or the GPL.
*
* ***** END LICENSE BLOCK ***** */

// AccessibarTTS.js is responsible for handling the "auditory" aspect of the toolbar
// on the Javascript side interacting with the Accessibar's Java Text To Speech reader
// implementation. It defines a Javascript object <code>AccessibarTTS</code> which
// is generally responsible for adding event listeners to the elements we wish to have
// read out loud. The AccessibarTTS object(whose creation is to be managed by
// <code>ttsLaunchManager()</code>) holds a reference to an instance of Accessibars Java
// Text To Speech reader singleton service <code>Packages.AccessibarReader.INSTANCE</code>.
// The relevant strings (and their context) are sent to the Java reader in separate
// threads to be read out loud.

/**
* Constructs the AccessibarTTS Object
* @constructor
*/
function AccessibarTTS() {
    // **** *ENFORCED* PRIVATE members thanks to JavaScript closures ****
    // (Inner functions have access to the vars and parameters of its outer function,
    // even after the outer function has returned.)
    /**
    * A reference to the Java AcceessibarReader Singleton service.
    * @private
    */
    var javaReader = Packages.AccessibarReader.INSTANCE;
    /**
    * A reference to the XPCOM nsIAccessibilityService
    * @private
    */
    var accService;
    /**
    * Boolean flag indicating that the reader has listeners listening to events.
    * @private
    */
    var isActive = false;

    // **** "Privileged" accessors ****
    // Enable access to the object's private members needed by its 'prototype' methods

    /** Java reader service accessor */
    this.getJavaReader = function() { return javaReader; };
    /** Accessibility service accessor */
    this.getAccService = function() { return accService; };
    /** Accessibility service setter */
    this.setAccService = function(newAccService) { accService = newAccService; };
    /** isActive getter */
    this.getIsActive = function() { return isActive; };
    /** isActive setter */
    this.setIsActive = function(aIsActive) { isActive = aIsActive; };
    /** Initialize this object */
    this.init();
}

// ** "STATIC" MEMBERS... **
/**
* Public static member will hold a reference to this object
*/
AccessibarTTS.thisObject = null;

/**
* Static reference to <code>handleAccessible()</code> event listener.
* Used when adding event listeners so that:
* <pre>
* 1) The event listener will be a method of THIS object rather than the element the event
*    is evoked on.
* 2) When removing the listener, the SAME method reference will be dereferenced..
*    Otherwise we won't be able to remove it as an event listener...
* </pre>
*/
AccessibarTTS.handleAccCallRef =
function(event){
    AccessibarTTS.thisObject.handleAccessible(event);
};

/**
* Static reference to <code>handleRegularText()</code> event listener. Needed for the same
* reason described for the static member above.
*/
AccessibarTTS.handleRegTextCallRef =
function(event){
    AccessibarTTS.thisObject.handleRegularText(event);
};

/**
* Static reference to <code>handleLoad()</code> event listener.
*/
AccessibarTTS.handleLoadCallRef =
function(event){
    AccessibarTTS.thisObject.handleLoad(event);
};

/**
* Static reference to the shutdown handler.
*/
AccessibarTTS.handleShutdownCallRef =
function(){
    AccessibarTTS.thisObject.handleShutdown();
};

/**
* These are Microsoft Accessibility Architecture State Flags Supported by Mozilla's
* nsAccessible. Further documentation can be found at the nsAccessible IDL:
*    http://lxr.mozilla.org/seamonkey/source/accessible/public/nsIAccessible.idl
*
* MSAA State flags - used for bitfield. More than 1 allowed.
* These will be utilized to obtain additional context data from the various events.
* This is important for blind individuals that need to know what the context of the
* string read out is.
* (*) At this point Accessibar indicates whether an element read out loud is a link
*     and can be extended to support additional context info to be read out.
*/
AccessibarTTS.stateDesc = [
"unavailable",
"selected",
"focused",
"pressed",
"checked",
"mixed",
"readonly",
"hottracked",
"default",
"expanded",
"collapsed",
"busy",
"floating",
"marqueed",
"animated",
"invisible",
"offscreen",
"sizeable",
"moveable",
"selfvoicing",
"focusable",
"selectable",
"linked",
"traversed",
"multiselectable",
"extselectable",
"alert_low",
"alert_medium",
"alert_high",
"protected",
"haspopup"
];


/**
* Accessibar's public methods defined within the object's prototype property.
*/
AccessibarTTS.prototype =
{
    /**
    * Initializes the AccessibarTTS object
    */
    init : function()
    {
        try {
            var accService = Components.classes["@mozilla.org/accessibilityService;1"]
                             .createInstance();
            accService = accService
                         .QueryInterface(Components.interfaces.nsIAccessibilityService);
            this.setAccService(accService);
        }
        catch (ex) {
            alert("** Error getting accessibility service **");
        }
        // Set out static reference to this object
        AccessibarTTS.thisObject = this;

        // Properly set voice settings from user's preferences
        this.setVoiceFromOptions();

        this.addListeners();
        // This is needed for the first time..
        this.addTextListenersToDocs();
    },

    /**
    * Sets voice settings from the user's preferences.
    */
    setVoiceFromOptions : function()
    {
        var prefService = Components.classes["@mozilla.org/preferences-service;1"]
                          .getService(Components.interfaces.nsIPrefService).getBranch("");
        // The following variables will hold the relevant value in the prefService.
        var voice;
        var volume;
        var speed;
        var pitch;
        var range;

        try {
            voice = prefService.getCharPref("accessibar.TextToSpeech.voice");
            volume = prefService.getIntPref("accessibar.TextToSpeech.volume");
            speed = prefService.getIntPref("accessibar.TextToSpeech.speed");
            pitch = prefService.getIntPref("accessibar.TextToSpeech.pitch");
            range = prefService.getIntPref("accessibar.TextToSpeech.range");

            this.getJavaReader().setReaderOptions(voice, volume, speed, pitch, range);
        } catch(e) {
          /*
            alert("Accessibar failed accessing user voice preferences.\n" +
                  "Default values will be used instead.");
          */
            alert("Oops! You seem to have forgotten to install Accessibar's TTS.\n" +
                  "Please visit http://accessibar.mozdev.org/installation.html\n" +
                  "for instructions on installing Accessibar Text To Speech.");               
        }
    },

    /**
    * Adds all required event listeners and sets <code>AccessibarTTS</code>'s
    * <code>isActive</code> to true.
    */
    addListeners : function()
    {
        // Accessibles Handler
        //function(event){ AccessibarTTS.thisObject.handleAccessible(event); },
        window.addEventListener("mouseover", AccessibarTTS.handleAccCallRef, true);

        //function(event){ AccessibarTTS.thisObject.handleAccessible(event); },
        window.addEventListener("focus", AccessibarTTS.handleAccCallRef, true);

        // Handler for Text nodes which have no Accessibles is a bit more involved...
        // On each page load we will add EventListeners directly to all relevant text
        // node's <span> parent which will be created.
        //function(event){ AccessibarTTS.thisObject.handleLoad(event); },
        window.addEventListener("load", AccessibarTTS.handleLoadCallRef, true);

        window.addEventListener("close", AccessibarTTS.handleShutdownCallRef, false);

        this.setIsActive( true );
    },

    /**
    * Removes all event listeners and flags the AccessibarTTS object as not active.
    */
    removeListeners : function()
    {
        //function(event){ AccessibarTTS.thisObject.handleAccessible(event); },
        window.removeEventListener("mouseover", AccessibarTTS.handleAccCallRef, true);

        //function(event){ AccessibarTTS.thisObject.handleAccessible(event); },
        window.removeEventListener("focus", AccessibarTTS.handleAccCallRef, true);
        //function(event){ AccessibarTTS.thisObject.handleLoad(event); },
        window.removeEventListener("load", AccessibarTTS.handleLoadCallRef, true);

        // Will remove document text listeners as well (quicker than traversing doc
        // and individually removing each listener on text node span parents)
        BrowserReload();

        this.setIsActive( false );
        
        // Stop reader if it is still talking
        this.getJavaReader().stop();
    },

    /**
    * Dispatches a thread to read out an accessible object
    */
    handleAccessible : function(event)
    {
        // START A THREAD>  THIS HELPS UI Responsiveness..
        window.setTimeout( function(){ AccessibarTTS.thisObject.readAccessible(event); }, 0);
    },
    
    /**
    * Dispatch a thread to read out a regular text string
    */ 
    handleRegularText : function(event)
    {
        window.setTimeout( function(){ AccessibarTTS.thisObject.readRegText(event); }, 0);
    },

    /**
    * This method detects what type of accessible node this is and calls
    * <code>AccessibarReader</code>'s <code>speak()</code> method to read it out.
    * If the target of the event is an accessible link node it is read out as a link.
    * @event The event whose target is an accessible node aquired from the XPCOM nsIAccessible
    *        service.
    */
    readAccessible : function(event)
    {
        var domNode = event.target;	// DOM node target of the event
        var accNode = null;             // nsIAccessible object
        var rawText = event.explicitOriginalTarget.textContent;  // "Inaccessible" raw text
        var readme = " "; 		// Extracted text from accessible node to be read

        try {
            accNode = this.getAccService().getAccessibleFor(domNode);

            // Filter out empty strings...
            if (!this.isReadableStr(accNode.name)) {
               /*
               // The following could be done but on condition that this domNode isn't a parent
               // of the WHOLE Document or Many other children..
               // accNode.name is null or all ws... BUT DOM Node HAS TEXT to be read!
               if ( this.isReadableStr(rawText) ) {
                   // Reformat characters that will not be properly handled
                   rawText = this.reformatStr( rawText );
                   this.getJavaReader().speak( rawText );
               }
               */
               return;
            }
            // Remove or replace all characters that will not be properly handled
            readme = this.reformatStr( accNode.name );

            // I have an accessible node, now see what type it is so that additional
            // context information can be read out loud..
            const stateStr = this.stateToString(accNode);

            // It's a link
            if ( stateStr.indexOf("linked") != -1 ) {
                this.getJavaReader().speak( readme + " link");
            }
            // Add checks for: It's a button?? ...It's a menu item...
            // Its a regular string
            else {
                this.getJavaReader().speak( readme );
            }
            //java.lang.System.out.println( "ACCESSIBLE NAME>> " + accNode.name );
            //java.lang.System.out.println( "ACCESSIBLE STATE>> " + accNode.state );
         } catch(ex) { }
    },

    /**
    * Adapted from Aaron Leventhal's "nsAccessibile from javascript" example
    * http://www.mozilla.org/access/samples/js-nsIAccessible
    *
    * Will give us extra state info of an nsIAccessible element we have managed to obtain
    * @param accNode an accessible node obtained from the nsIAccessible service.
    */
    stateToString : function (accNode) {
        var state = accNode.state;
        var stateStr = "";
        var states = new Array();
        var i;
        for (i = 0; i < AccessibarTTS.stateDesc.length; ++i) {
            if (state & (1 << i)) {
                states.push(AccessibarTTS.stateDesc[i]);
            }
        }
        if (states.length) {
            stateStr = " ( " + states.join(" | ") + " ) ";
        }
        return stateStr;
    },

    /**
    * Reads the text content of an event
    * @param event The event whose text is to be read out.
    */
    readRegText : function( event )
    {
        var readme = event.explicitOriginalTarget.textContent; 

        // Filter out whitespace or null text nodes...
        // Also ignore those calls made after flagged as inactive and before browser
        // fully reloads effectively removing regular text event listeners.
        if ( !this.isReadableStr( readme ) || !this.getIsActive() ) {
            return;
        }
        
        // Normalize characters that will not be properly handled
        readme = this.reformatStr( readme );

        this.getJavaReader().speak( readme );

    },

    /**
    * Utility method to acertain whether a given string is not all whitespaces or null.
    * @param string The string to test
    * @return true if the string is readable and false otherwise.
    */
    isReadableStr : function ( string ) {
        if (string == null || !(/[^\t\n\r\s]/.test(string))) {
            return false;
        }
        return true;
    },

    /**
     * This is a Hack/workaround to resolve an enigmatic runtime issue:
     * A nasty runtime bug which caused the reader to stop, followed by browser instability.
     * reformatStr leaves us with characters that are sure not to cause problems.
     * Since a subset of "alphanumerical and some punctuation" chars has resolved this issue,
     * and still allows the reader to speak perfect English, this will enforce the
     * charater subset used.
     * (Be careful when allowing additional chars... Even adding a Period will down the
     * reader.)
     * @param string String to reformat before being read by Java reader
     * @return A string assured to be OK to send to Java
     */
    reformatStr : function ( string ) {
        // badChars are:
        //(not alphanumeric or the puncs listed *period purposely left out*)
        const badChars =  /[^a-zA-Z0-9,@%]/g;
        const ampersand = /&/g;
        const period = /\./g;
        const apostrophe = /'/g;

        // Note, Reader also falls if "." is replaced with a simalar pause sequence: " ' " so
        // I am left with doing the following:
        string = string.replace( period, ", " );   // Ugh..
        // Avoid " ' " sequence.
        // (TODO: do this smarter on Java side.)
        string = string.replace( apostrophe, "");

        // "&" should be read "and" rather that the voice engine's default "ampersand"
        string = string.replace( ampersand, " and " );
        string = string.replace( badChars, " " );
     
        return string;
    },

    /**
    * Event handler for "load" events which will add text listener's to a newly
    * loaded web document.
    * @param event The load event to handle
    */
    handleLoad : function(event)
    {
        // We only wish to handle document page loads
        if(!event.originalTarget == content.document) {
            return;
        }
        this.addTextListenersToDocs();
    },

    /**
    * Takes care of adding text listeners to all the loaded document frames.
    * This is done by obtaining an Array of all the documents and then calling
    * the <code>addListenersToText</code> method for each document frame.
    */
    addTextListenersToDocs : function()
    {
        var docsList = xsiGetAllDocs(window.content, new Array());
        for (var i=0; i < docsList.length; ++i) {
            // links are nsIAccessible's so we get it from that
            //        this.addListenerToLinks(docsList[i]);
            this.addListenersToText(docsList[i].body);
        }
    },

    /**
    * Adds listener to all child text nodes of a given parent node. This is done by
    * reparenting the text node with a "SPAN" element to which a listener can be added
    * to. (A text node on its own cannot have an event listener registered to it).
    * @param node A parent node whose children text nodes are to be reparented with a
    *        span node which will have the event listener added to.
    */
    addListenersToText : function(node)
    {
        var accNode = null;
        // Check if node is a valid Text object
        // We only want to do this once for any given text node
        if (xsiIsValidTextNode(node) &&
        node.parentNode.getAttribute("class") != "accessibar-span-text") {
            // Accessibility API provides no other means of discovering
            // whether a text node has an accessible parent..
            // (@@ This code doesn't scale well for semi large documents <500kb @@)
            ///  Make a list of all accessible nodes and check whether it's on the
            // list instead of the following...
            try {
                accNode = this.getAccService().getAccessibleFor(node.parentNode);
            } catch( e ) { }
            // Node is accessible no need to reparent
            if ( accNode ) {
                return node;
            }

            var newParent = xsiReparentNode( node, "span" );
            newParent.addEventListener("mouseover",
                                       AccessibarTTS.handleRegTextCallRef,
                                       true);
            newParent.setAttribute("class", "accessibar-span-text");
            // return, since text nodes have no children
            // with the newParent which has replaced (fathered) text "node"
            return newParent;

        }
        // Otherwise, n may have children whose characters we need to count
        // Instead of using the childNodes property, this loop examines the
        // children of n using the firstChild and nextSibling properties.
        for(var m = node.firstChild; m != null; m = m.nextSibling) {
            // We are parenting text nodes to span tags so recursion should continue
            // from parent nodes of such text nodes...
            m = this.addListenersToText(m);
        }
        return node;
    },

    /**
    * Deallocate resources aquired for the Java reader service.
    */
    handleShutdown : function() {
        this.getJavaReader().close();
    }
};

/**
* Responsible for creating AccessibarTTS object when necessary. It will be used each time
* Reader On will be selected.
* @param onState Boolean flag indicating the state desired: true - on, false - off
*/
function ttsLaunchManager( onState ) {
    // Reader On request:
    if ( onState ) {
        // Create new initial object
        if (!AccessibarTTS.thisObject) {
            // UI Responsiveness is the way to go... dispatch thread
            window.setTimeout("AccessibarTTS.thisObject = new AccessibarTTS();", 0);
        }
        // AccessibarTTS already exists so just restore listeners once again
        else if (!AccessibarTTS.thisObject.getIsActive()) {
            AccessibarTTS.thisObject.addListeners();
            AccessibarTTS.thisObject.addTextListenersToDocs();
        }
    }
    // Off request...
    else {
        if (AccessibarTTS.thisObject && AccessibarTTS.thisObject.getIsActive() )
        AccessibarTTS.thisObject.removeListeners();
    }
}


